﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Syncfusion.Windows.Controls.Grid;
using System.Runtime.Serialization;
using System.Windows;
using System.Windows.Input;
using Syncfusion.Windows.Controls.Cells;
using System.Windows.Media;
using System.Collections;
using System.ComponentModel;
using System.Windows.Data;

namespace WpfApplication1
{
    class DragNodeHelper<T>
    {
        GridTreeControl tree = null;
        
        bool isDragging = false;
        RowColumnIndex draggedRow = RowColumnIndex.Empty;
        RowColumnIndex targetRow = RowColumnIndex.Empty;
        Pen draggedPen = new Pen(Brushes.LightGray, 4);
        Pen targetPen = new Pen(Brushes.Red, 4);
        bool wasExpanded = false;

        public DragNodeHelper(GridTreeControl tree)
        {
            this.tree = tree;
            //   tree.ShowRowHeader = true; //use headers to distinguish between dragging and row selects 
            tree.PreviewMouseLeftButtonDown += tree_PreviewMouseLeftButtonDown;
            tree.PreviewMouseLeftButtonUp += tree_PreviewMouseLeftButtonUp;
            tree.InternalGrid.PrepareRenderCell += InternalGrid_PrepareRenderCell;
            tree.PreviewKeyDown += tree_PreviewKeyDown;
            tree.PreviewMouseMove += tree_PreviewMouseMove;
        }

        private void tree_PreviewMouseMove(object sender, MouseEventArgs e)
        {
            if (Mouse.LeftButton == MouseButtonState.Released)
            {
                CancelDrag();
                return;
            }
            if (isDragging)
            {

                Point p1 = e.GetPosition(tree.InternalGrid);
                RowColumnIndex rc = tree.InternalGrid.PointToCellRowColumnIndex(p1);
                if (rc.RowIndex == 0)
                    return;
                if (rc.RowIndex != targetRow.RowIndex)
                {
                    ClearTargetQueue();
                    targetRow = rc;
                    DrawDropQueues();
                }
                if (rc.RowIndex >= tree.InternalGrid.ScrollRows.LastBodyVisibleLineIndex - 1 &&
                    rc.RowIndex < tree.InternalGrid.Model.RowCount)
                {
                    tree.InternalGrid.ScrollInView(new RowColumnIndex(rc.RowIndex + 1, 1));
                    System.Threading.Thread.Sleep(100);
                }
                if (rc.RowIndex <= tree.InternalGrid.ScrollRows.ScrollLineIndex + 1 && rc.RowIndex > 2)
                {
                    tree.InternalGrid.ScrollInView(new RowColumnIndex(rc.RowIndex - 1, 1));
                    System.Threading.Thread.Sleep(100);
                }

                tree.InternalGrid.InvalidateCell(GridRangeInfo.Col(1));
                tree.InternalGrid.InvalidateVisual();
            }
        }

        private void CancelDrag()
        {
            isDragging = false;
            ClearDropQueues();
            draggedRow = RowColumnIndex.Empty;
            targetRow = RowColumnIndex.Empty;
        }

        private void tree_PreviewKeyDown(object sender, KeyEventArgs e)
        {
            if (isDragging && e.Key == Key.Escape)
            {
                isDragging = false;
                int rowIndex = draggedRow.RowIndex;
                ClearDropQueues();
                if (wasExpanded)
                {
                    tree.ExpandNode(rowIndex);
                }
            }
        }

        private void InternalGrid_PrepareRenderCell(object sender, GridPrepareRenderCellEventArgs e)
        {
            if (isDragging)
            {
                if (e.Cell.RowIndex == draggedRow.RowIndex && e.Cell.ColumnIndex == 1)
                {
                    e.Style.Borders.All = draggedPen;
                }
                else if (e.Cell.RowIndex == targetRow.RowIndex && e.Cell.ColumnIndex == 1)
                {
                    e.Style.Borders.Top = targetPen;
                }
                //  Console.WriteLine("{0} {1}", draggedRow.RowIndex, targetRow.RowIndex);
            }
        }

        private void tree_PreviewMouseLeftButtonUp(object sender, MouseButtonEventArgs e)
        {
            isDragging = false;
            int fromRow = draggedRow.RowIndex;
            int toRow = targetRow.RowIndex;
            ClearDropQueues();
            if (fromRow != RowColumnIndex.Empty.RowIndex && toRow != RowColumnIndex.Empty.RowIndex)
            {
                MoveRow(fromRow, toRow);

                tree.Dispatcher.BeginInvoke(new Action(() =>
                    {
                        if (toRow > fromRow && toRow > 1)
                            toRow--;

                        if (tree.InternalGrid.CurrentCell.HasCurrentCell)
                        {
                            tree.InternalGrid.CurrentCell.MoveTo(toRow, tree.InternalGrid.CurrentCell.ColumnIndex);
                        }
                        else
                        {
                            tree.SelectedNodes.SetSelected(tree.InternalGrid.GetNodeAtRowIndex(toRow), true);
                        }

                    }), null);
            }
        }

        protected virtual void MoveRow(int fromRowIndex, int toRowIndex)
        {
            if (toRowIndex == fromRowIndex)
                return;

            var draggedRowNode = tree.InternalGrid.GetNodeAtRowIndex(fromRowIndex);
            var targetRowNode = tree.InternalGrid.GetNodeAtRowIndex(toRowIndex);

            //do the remove ........................
            T itemToRemove = (T) draggedRowNode.Item;

            ////collapse parent node, remove the item, and then expand again
            //tree.CollapseNode(draggedRowNode.ParentNode);
            ////get the parentList so item can be removed from it...
            PropertyDescriptorCollection pdc = TypeDescriptor.GetProperties(typeof (T));
            IList<T> parentList = null;
            if (draggedRowNode.ParentNode == null) //root
            {
                parentList = tree.ItemsSource as IList<T>;

            }
            else
            {
                parentList = pdc[tree.ChildPropertyName].GetValue(draggedRowNode.ParentNode.Item) as IList<T>;
            }
            parentList.Remove((T) draggedRowNode.Item); //remove the item        

            //do the insert ...............................
            int loc;
            if (targetRowNode.ParentNode == null)
            {
                parentList = tree.ItemsSource as IList<T>;
                loc = tree.InternalGrid.RootNodes.IndexOf(targetRowNode);
                parentList.Insert(loc, (T) draggedRowNode.Item);
            }
            else
            {
                loc = targetRowNode.ParentNode.ChildNodes.IndexOf(targetRowNode);
                parentList = pdc[tree.ChildPropertyName].GetValue(targetRowNode.ParentNode.Item) as IList<T>;
                parentList.Insert(loc, (T) draggedRowNode.Item);
            }
            wasDropped = true;
        }

        private void ChangeAllNodeLevels(GridTreeNode draggedRowNode, int level)
        {
            draggedRowNode.Level = level;
            foreach (GridTreeNode n in draggedRowNode.ChildNodes)
            {
                ChangeAllNodeLevels(n, level + 1);
            }
        }

        bool wasDropped = false;

        void tree_PreviewMouseLeftButtonDown(object sender, MouseButtonEventArgs e)
        {
            Point p1 = e.GetPosition(tree.InternalGrid);
            RowColumnIndex rc = tree.InternalGrid.PointToCellRowColumnIndex(p1);
            
            GridTreeNode node = tree.InternalGrid.GetNodeAtRowIndex(rc.RowIndex);
            if (rc.RowIndex > 1)
            {
                if (rc.ColumnIndex == 1 && p1.X < 20)
                {
                    //over glyph
                    return;
                }
                isDragging = true;
                draggedRow = rc;
                wasExpanded = node.Expanded;
                if (wasExpanded)
                {
                    tree.CollapseNode(draggedRow.RowIndex);

                    tree.SelectedNodes.SetSelected(node, true);
                }
                if (!tree.IsKeyboardFocusWithin)
                    tree.Focus();
                e.Handled = true;
                DrawDropQueues();
            }
            else if (wasDropped && Keyboard.Modifiers == ModifierKeys.None)
            {
                tree.SelectedNodes.Clear();
            }
            wasDropped = false;
        }


        private void ClearTargetQueue()
        {
            RowColumnIndex oldTar = targetRow;
            targetRow = RowColumnIndex.Empty;
            tree.InternalGrid.InvalidateCell(GridRangeInfo.Cell(oldTar.RowIndex, 0));
        }

        private void ClearDropQueues()
        {
            var oldDrag = draggedRow;
            var oldTar = targetRow;
            draggedRow = RowColumnIndex.Empty;
            targetRow = RowColumnIndex.Empty;
            tree.InternalGrid.InvalidateCell(GridRangeInfo.Cell(oldDrag.RowIndex, 0));
            tree.InternalGrid.InvalidateCell(GridRangeInfo.Cell(oldTar.RowIndex, 0));
            tree.InternalGrid.InvalidateVisual();
        }

        private void DrawDropQueues()
        {
            if (draggedRow != RowColumnIndex.Empty)
            {
                tree.InternalGrid.InvalidateCell(GridRangeInfo.Cell(draggedRow.RowIndex, 0));
            }
            if (targetRow != RowColumnIndex.Empty)
            {
                tree.InternalGrid.InvalidateCell(GridRangeInfo.Cell(targetRow.RowIndex, 0));
            }
        }
    }
}
